import 'dart:typed_data';

import '../utils/Converter.dart';
import 'Stream.dart';
import 'Tag.dart';
import 'impl/CompoundTag.dart';

class NbtIo {
  static ByteLayer _io = ByteLayer();

  // Handle various helper functions here!

  static Future<void> _read(String file,
      [void Function(int)? onReadListener]) async {
    _io = ByteLayer();
    _io.addListener(onReadListener);

    await _io.readFromFile(file);
  }

  // This function will read the file and check if it is infact gzipped
  static Future<Tag> read(String file,
      [void Function(int)? onReadListener]) async {
    _io.disposeStream();
    await _read(file, onReadListener);
    if (_io.readByte() == TagType.Compound.byte) {
      _io.resetPosition();
      return Tag.readNamedTag(_io);
    } else {
      // Is likely gzip compressed
      return readCompressed(file);
    }
  }

  static Future<Tag> readCompressed(String file,
      [void Function(int)? onReadListener]) async {
    _io.disposeStream();
    _io = ByteLayer();
    _io.addListener(onReadListener);
    await _io.readFromFile(file);
    await _io.decompress();
    _io.resetPosition();
    return Tag.readNamedTag(_io);
  }

  static Future<void> write(String file, CompoundTag tag,
      [void Function(int)? onWriteListener]) async {
    _io.disposeStream();
    _io = ByteLayer();
    _io.addListener(onWriteListener);

    Tag.writeNamedTag(tag, _io);
    await _io.writeToFile(file);
  }

  static Future<void> writeCompressed(String file, CompoundTag tag,
      [void Function(int)? onWriteListener]) async {
    _io.disposeStream();
    _io = ByteLayer();
    _io.addListener(onWriteListener);
    Tag.writeNamedTag(tag, _io);
    await _io.compress();
    await _io.writeToFile(file);
  }

  static ByteLayer getStream() {
    return _io;
  }

  static Stream<int> getByteStream() {
    return _io.getStream();
  }

  static Future<String> writeBase64StringCompressed(CompoundTag tag,
      [void Function(int)? onWriteListener]) async {
    _io.disposeStream();
    _io = ByteLayer();
    _io.addListener(onWriteListener);

    Tag.writeNamedTag(tag, _io);
    _io.compress();
    return base64Encoder.encode(_io.bytes);
  }

  static Future<Tag> readBase64StringCompressed(String encoded,
      [void Function(int)? onReadListener]) async {
    _io.disposeStream();
    _io = ByteLayer();
    _io.addListener(onReadListener);

    List<int> bytes = base64Encoder.decode(encoded);
    for (int byte in bytes) {
      _io.writeByte(byte);
    }
    _io.decompress();
    _io.resetPosition();
    return Tag.readNamedTag(_io);
  }

  static Future<String> writeBase64String(CompoundTag tag,
      [void Function(int)? onWriteListener]) async {
    _io.disposeStream();
    _io = ByteLayer();
    _io.addListener(onWriteListener);
    Tag.writeNamedTag(tag, _io);
    return base64Encoder.encode(_io.bytes);
  }

  static Future<Tag> readBase64String(String encoded,
      [void Function(int)? onReadListener]) async {
    _io.disposeStream();
    _io = ByteLayer();
    _io.addListener(onReadListener);

    List<int> bytes = base64Encoder.decode(encoded);
    for (int byte in bytes) {
      _io.writeByte(byte);
    }
    _io.resetPosition();
    return Tag.readNamedTag(_io);
  }

  /// Writes the NBT Data to stream in compressed mode
  static Future<Uint8List> writeToStreamCompressed(CompoundTag tag,
      [void Function(int)? onWriteListener]) async {
    _io.disposeStream();
    _io = ByteLayer();
    _io.addListener(onWriteListener);

    Tag.writeNamedTag(tag, _io);
    _io.compress();

    return _io.bytes;
  }

  static Future<Tag> readFromStreamCompressed(Uint8List list,
      [void Function(int)? onReadListener]) async {
    _io.disposeStream();
    _io = ByteLayer();
    _io.addListener(onReadListener);
    try {
      _io.writeBytes(list);
      _io.resetPosition();
      _io.decompress();
      _io.resetPosition();
    } catch (E) {
      print(E);
    } finally {
      return Tag.readNamedTag(_io);
    }
  }

  static Future<Uint8List> writeToStream(CompoundTag tag,
      [void Function(int)? onWriteListener]) async {
    _io.disposeStream();
    _io = ByteLayer();
    _io.addListener(onWriteListener);

    Tag.writeNamedTag(tag, _io);

    return _io.bytes;
  }

  static Future<Tag> readFromStream(Uint8List list,
      [void Function(int)? onReadListener]) async {
    _io.disposeStream();
    _io = ByteLayer();
    _io.addListener(onReadListener);

    try {
      _io.writeBytes(list);
      _io.resetPosition();
    } catch (E) {
      print(E);
    } finally {
      if (_io.readByte() == TagType.Compound.byte) {
        _io.resetPosition();
        return Tag.readNamedTag(_io, depth: 0);
      } else {
        _io.resetPosition();
        _io.decompress();
        _io.resetPosition();
        return Tag.readNamedTag(_io, depth: 0);
      }
    }
  }
}
